"""
This example demonstrates how to use the AqVISALibrary class to select the
application type, open the resource manager, write a command, read the
response, and close the resource manager.
"""

import sys
from aqvisa import AqVISALibrary
from aqvisa import AppType

# Initialize the AqVISALibrary
manager = AqVISALibrary()

# Select the application type
success = manager.select_app_type(AppType.MIXEDSIGNALOSCILLOSCOPE)

if not success:
    print(f"viSelectAppType failed: {success}")
    sys.exit(1)

# Connect to the selected application
success = manager.open()
if not success:
    print(f"viOpenRM failed: {success}")
    sys.exit(1)

# Send the command `*IDN?` to the application
print("viWrite *IDN?")
success = manager.write(command=b"*IDN?")
if not success:
    print(f"viWrite failed, error code: {manager.get_status_code()}")

# Read the response
# The response of `*IDN?` command is like
# Model: <Model Name>, Serial No.: <Device Serial Number>;
# If the application is in demo mode, the response will be like
# Model: Demo, Serial No.: Demo;
response = manager.read(count=1024)
print(f"viRead Response: {response}")

# Disconnect from the application
_ = manager.close()



"""
This module defines the supported application type of Acute's application
for the AqVISA library.
"""

from enum import IntEnum


class AppType(IntEnum):
    """
    This enum defines the supported application type of Acute's application.
    """

    # TravelLogic Application
    TRAVELLOGIC = 0

    # BusFinder & Logic Analyzer Application
    BUSFINDER_LOGICANALYZER = 1

    # TravelBus Application
    TRAVELBUS = 2

    # Mixed Signal Oscilloscope (MSO) Application
    MIXEDSIGNALOSCILLOSCOPE = 3

    # Digital Storage Oscilloscope (DSO) Application
    DIGTIALSTORAGEOSCILLOSCOPE = 101



"""
Include all classes and functions in the aqvisa package.
"""

from aqvisa.apptype import AppType
from aqvisa.aqvisalibrary import AqVISALibrary
from aqvisa.status_code import StatusCode

__all__ = ["AppType", "AqVISALibrary", "StatusCode"]



"""
Cross platform helper for ctypes interface
"""

import ctypes
import sys

if sys.platform == 'win32':
    C_FUNCTYPE, C_DLL = ctypes.WINFUNCTYPE, ctypes.WinDLL
else:
    C_FUNCTYPE, C_DLL = ctypes.CFUNCTYPE, ctypes.CDLL

__all__ = ["C_FUNCTYPE", "C_DLL"]



"""
This module defines the AqVISALibrary class.

AqVISALibrary is a Python wrapper for AqVISA / AqVISA dynamic library,
which simplifies the access to the AqVISA functions from Python.
"""

from ctypes import c_int, c_char_p, create_string_buffer
from aqvisa.apptype import AppType
from aqvisa.status_code import StatusCode
from aqvisa.utils.library_wrapper import AqLibrary


class AqVISALibrary(AqLibrary):
    """
    Main class for AqVISALibrary to access Acute's instruments.

    Creating AqVISALibrary object will attempt to load the AqVISA library.
    """

    def __init__(self) -> None:
        """
        Initialize the AqVISALibrary object and resolves the methods from
        the library.

        Here we only handle AqVISA64 (64-bit on Windows).

        TODO: Support other platforms.
        """
        super().__init__("AqVISA64")
        self._functions = [
            ["_viCloseRM", "viCloseRM", c_int, None],
            ["_viErrCode", "viErrCode", c_int, None],
            ["_viGetCommandResult", "viGetCommandResult", c_int, None],
            ["_viOpenRM", "viOpenRM", c_int, None],
            ["_viRead", "viRead", c_int, [c_char_p, c_int]],
            ["_viSelectAppType", "viSelectAppType", c_int, [c_int]],
            ["_viWrite", "viWrite", c_int, [c_char_p]],
        ]

        for method in self._functions:
            self.map_api(*method)

    def close(self) -> int:
        """
        Disconnect from the application.

        This function corresponds to viCloseRM function
        of the AqVISA library.

        :return: 1 on success, otherwise 0.
        """
        return getattr(self, "_viCloseRM")()

    def get_status_code(self) -> StatusCode:
        """
        Retrieve the AqVISA latest status code.

        This function corresponds to viErrCode function
        of the AqVISA library.

        :return: AqVISA status code.
        """
        status = getattr(self, "_viErrCode")()
        return StatusCode(status)

    def get_command_result(self) -> int:
        """
        Get the execution result after the command is executed.

        This function corresponds to viGetCommandResult function
        of the AqVISA library.

        :return: The result after the command is executed. If the command does
                 not return any data, it will return an empty bytes object.
        """
        return getattr(self, "_viGetCommandResult")()

    def open(self) -> int:
        """
        Connect to the application.

        This function corresponds to viOpenRM function of the AqVISA library.

        :return: 1 on success, otherwise 0.
        """
        return getattr(self, "_viOpenRM")()

    def read(self, count: int) -> bytes:
        """
        Get the result after the command is executed.

        This function corresponds to viRead function
        of the AqVISA library.

        :param count: The number of bytes to read.
        :return: The result after the command is executed. If the command does
                 not return any data, it will return an empty bytes object.
        """

        buffer = create_string_buffer(count)
        ret_size = getattr(self, "_viRead")(buffer, count)
        return buffer.value if ret_size else b''

    def select_app_type(self, app_type: AppType) -> int:
        """
        Setup the target connection application.

        This function corresponds to viSelectAppType function
        of the AqVISA library.

        :param app_type: The application type to select.
        :return: 1 on success, otherwise 0.
        """
        return getattr(self, "_viSelectAppType")(app_type.value)

    def write(self, command: bytes) -> int:
        """
        Send AqLAVISA command to the application.

        This function corresponds to viWrite function
        of the AqVISA library.

        :param command: The command to send to the application.
        :return: The number of bytes written to the application.
        """
        return getattr(self, "_viWrite")(command)



"""
This is a wrapper for the AqLibrary class.
"""

import time
from ctypes.util import find_library
from typing import Any, List, Optional
from aqvisa.utils.ctypes_wrapper import C_DLL


class AqLibrary:
    """
    This is a wrapper for the AqLibrary class.
    """
    def __init__(self, libname: str) -> None:
        self.libname = libname
        self._load()

    def __str__(self) -> str:
        return f"Acute {self.libname} library"

    def _load(self) -> Any:
        lib_path = find_library(self.libname)

        # Check if library exists
        if lib_path is None:
            raise FileNotFoundError(f"Library {self.libname} not found.")

        self._dll = C_DLL(lib_path)
        time.sleep(0.05)

    def map_api(self,
                py_api_name: str,
                c_api_name: str,
                restype: Optional[Any] = None,
                argstype: Optional[List[Any]] = None) -> None:
        """
        This function maps the python APIs to the C APIs.
        """
        try:
            c_api = getattr(self._dll, c_api_name)
            c_api.argtypes = argstype
            c_api.restype = restype
            setattr(self, py_api_name, c_api)

        except AttributeError:
            print(f"Fail to create {py_api_name} mapping")



"""
This module contains the status code for the AqVISA library.
"""

from enum import IntEnum


class StatusCode(IntEnum):
    """
    This collection of constants defines the status codes / error codes for
    the AqVISA library.
    """

    # Operation succeeded without error
    AQVI_NO_ERROR = 0

    # Error: OPENRM should only be called once to establish the software
    # connection
    AQVI_ALREADY_OPEN = 1

    # Error: The main software application was not started before calling the
    # viOPENRM
    AQVI_APPLICATION_NOT_STARTED = 2

    # Error: Failed to establish the connection to software application
    AQVI_FAILED_TO_CONNECT = 3

    # Error: The software connection is not valid yet, call viOpenRM first to
    # establish the connection
    AQVI_CONNECTION_NOT_VALID = 4

    # Error: The software application has been closed or re-started, need to
    # call viCloseRM and viOpenRM to establish the connection again
    AQVI_CONNECTION_LOST = 5

    # Error: Unable to send data to software application
    AQVI_FAILED_TO_COMMUNICATE = 6

    # Error: There's no any data returned from the software application
    AQVI_NO_RETURN_DATA = 7

    # Error: Input data buffer is NULL pointer and not valid
    AQVI_INPUT_NULL_POINTER = 8

    # Error: Input data buffer size is too small
    AQVI_DATA_BUFFER_TOO_SMALL = 9

    # Error: Select Application type should be done before OpenVM
    AQVI_SEL_APP_AFTER_VM_OPENED = 10

    # Error: Selected invalid Application type
    AQVI_SEL_APP_TYPE_INVALID = 11

    # Error: Previous Input Command Still Processing
    AQVI_PREVIOUS_CMD_PROCESSING = 12

    # Error: Input parameter unknown or not supported
    AQVI_INPUT_PARAMETER_UNKNOWN = 13

    # Error: Input parameter incompleted
    AQVI_INPUT_PARAMETER_INCOMPLETED = 14

    # Error: Input request timeout
    AQVI_TIMEOUT = 15

    # Command Error: Not supported in current SW (ex. for TBA)
    AQVI_NOT_SUPPORTED = 1000

    # Command Error: Input command incomplete
    AQVI_INCOMPLETE_COMMAND = 1001

    # Command Error: Input command requires exist SubWindow
    AQVI_SUBWND_INVALID = 1002

    # Command Error: Input command ask to add new SubWindow page, but software
    # already exceed maximum page count
    AQVI_SUBWND_CNT_EXCEED = 1003

    # Command Error: Input command ignored while software busy
    AQVI_SW_BUSY = 1004

    # Command Error: Input command requires exist Logic Analyzer SubWindow
    AQVI_LASUBWND_INVALID = 1005

    # Command Error: Input command requires exist Protocol Analyzer SubWindow
    AQVI_PASUBWND_INVALID = 1006

    # Command Error: Input command requires exist Decode Report
    AQVI_DECODE_REPORT_INVALID = 1007

    # Command Error: Input commands requires exist Timing Report
    AQVI_TIMING_REPORT_INVALID = 1008

    # Command Error: Input command format error
    AQVI_INPUT_COMMAND_FORMAT_ERROR = 1009

    # Command Error: Input file directory invalid
    AQVI_INPUT_FILE_DIR_INVALID = 1010

    # Command Error: Sending Capture Start command while capture already in
    # progress
    AQVI_CAPTURE_ALREADY_RUNNING = 1011

    # Command Error: Sending Capture Stop command while capture is not running
    AQVI_CAPTURE_NOT_RUNNING = 1012

    # Command Error: Input Row or Column index invalid
    AQVI_ROW_COL_INDEX_INVALID = 1013

    # Command Error: Input index selection invalid
    AQVI_SELECT_INDEX_INVALID = 1014

    # Command Error: Missing input parameters or not enough input parameter
    AQVI_INPUT_PARAMETER_INVALID = 1015

    # Command Error: Input setting file format error
    AQVI_INPUT_SETTING_FILE_FORMAT_ERROR = 1016

    # Command Error: Unable to access File unable to reach select file
    # directory or selected unsupported file type
    AQVI_FILE_ACCESS_ERROR = 1017

    # Command Error: Input commands requires exist Transition Report
    AQVI_TRANSITION_REPORT_INVALID = 1018

    # Command Error: Input commands requires exist Measurement Report
    AQVI_MEASUREMENT_REPORT_INVALID = 1019

    # Command Error: Selected Label Name not valid
    AQVI_SELECT_LABEL_INVALID = 1020

    # Command Error: Selected Range Condition Error, can't mix Line and Cursor
    # Range Condition
    AQVI_SELECT_RANGE_ERROR = 1021

    # Command Error: Selected Channel Name in Cursor Search is not valid
    AQVI_SELECT_CHANNEL_NAME_INVALID = 1022

    # Command Error: Cursor Search Failed unable to find more valid match
    AQVI_CURSOR_SEARCH_FAILED = 1023

    # Command Error: Waveform Data Not Ready
    AQVI_WAVEFORM_NOT_READY = 1024

    # Command Error: software is not configured in EV mode
    AQVI_NOT_IN_EV_MODE = 1025

    # Command Error: software doesn't have any valid EV analysis result
    AQVI_NO_EV_ANALYSIS_RESULT = 1026

    # Command Error: selected Device S/N not presented in device list
    AQVI_SELECT_DEV_NOT_EXIST = 1027

    # Error: Unsupported feature, feature is not supported in current model or
    # demo mode
    AQVI_UNSUPPORT_FEATURE = 9995

    # Error: Unfinished Feature
    AQVI_UNFINISHED_FEATURE = 9996

    # Error: Unknown Command
    AQVI_UNKNOWN_COMMAND_ERROR = 9997

    # Error: Unknown Error
    AQVI_UNKNOWN_ERROR = 9998

    # Error: DLL initialization not finished, please retry again later
    AQVI_DLL_NOT_READY = 9999
